from majava.matchers import Matcher, Mismatch, _match_dict, make_matcher, _match, Absent


class DictContains(Matcher):
    """ If value contains expected key-value pairs
    """

    def __init__(self, expected, *, recursive=True):
        self.expected = expected
        self.recursive = recursive

    def __repr__(self):
        return f"DictContains({repr(self.expected)})"

    def _match(self, other):
        _match_dict(self.expected, other, allow_unexpected=True)


class IsInstance(Matcher):
    def __init__(self, *types):
        self.types = types

    def _types_str(self):
        return "|".join(i.__name__ for i in self.types)

    def __repr__(self):
        return f"IsInstance({self._types_str()})"

    def _match(self, other):
        if not isinstance(other, self.types):
            raise Mismatch(other, "", f"not {self}")


class Round(Matcher):
    def __init__(self, value, digits=0):
        self.value = round(value, digits)
        self.digits = digits

    def __repr__(self):
        return f"Round({self.value})"

    def _match(self, other):
        if self.value != round(other, self.digits):
            raise Mismatch(other, "", f"not ~{self.value}")


class ContainsOrdered(Matcher):
    def __init__(self, items):
        self.items = items

    def __repr__(self):
        return f"ContainsOrdered({self.items})"

    def _match(self, other):
        idx = 0
        for it in self.items:
            try:
                idx = other[idx:].index(it) + 1
            except ValueError:
                raise Mismatch(other, "", f"{repr(it)} is not in order")


class _Contains(Matcher):
    def __init__(self, items, ordered=False):
        self.items = items
        self.ordered = ordered

    def __repr__(self):
        return f"Contains({self.items})"

    def _match(self, other):
        missing_items = []
        for it in self.items:
            if it not in other:
                missing_items.append(it)

        if missing_items:
            raise Mismatch.missing_items(other, missing_items)


class Unordered(_Contains):
    """ Value must contains all expected items in any order
    """

    def __repr__(self):
        return f"Unordered({self.items})"

    def _match(self, other):
        if len(self.items) != len(other):
            raise Mismatch(other, "", f"len is not {len(self.items)}")
        super()._match(other)


def Contains(items, ordered=False):
    if ordered:
        return ContainsOrdered(items)
    return _Contains(items)


@make_matcher
def StartsWith(value, expected):
    return value.startswith(expected)


@make_matcher
def EndsWith(value, expected):
    return value.endswith(expected)


@make_matcher
def LengthIs(value, expected):
    return len(value) == expected


@make_matcher
def InInterval(value, low, high):
    return low <= value <= high


@make_matcher
def HasAttrs(value, **kwargs):
    for key, expected in kwargs.items():
        try:
            _match(expected, getattr(value, key, Absent))
        except Mismatch as e:
            raise e.prepend(key)
    return True



from .matchers import Matcher, Mismatch, _match, Any
import json


class IsJson(Matcher):
    def __init__(self, expected=Any):
        self.expected = expected

    def __repr__(self):
        return f"IsJson({repr(self.expected)})"

    def _match(self, other):
        try:
            other_obj = json.loads(other)
        except TypeError as e:
            raise Mismatch(other, "FromJSON", f"invalid type - {e}")
        except json.JSONDecodeError as e:
            raise Mismatch(other, "FromJSON", f"invalid JSON - {e}")

        try:
            _match(self.expected, other_obj)
        except Mismatch as e:
            raise e.prepend("FromJSON")



from .matchers import Matcher, Mismatch
import os


class _IsDirectory(Matcher):
    def __init__(self, is_empty=None):
        self._is_empty = is_empty

    def __repr__(self):
        if self._is_empty is not None:
            param = "empty" if self._is_empty else "not empty"
            return f"IsDirectory({param})"
        return "IsDirectory"

    def _match(self, other):
        try:
            if not os.path.isdir(other):
                raise Mismatch(other, "", "is not a directory")
        except TypeError as e:
            raise Mismatch(other, "", f"is not a directory (invalid type - {e})") from e

        if self._is_empty is not None:
            is_empty = not os.listdir(other)
            if self._is_empty and not is_empty:
                raise Mismatch(other, "", "directory is not empty")
            if not self._is_empty and is_empty:
                raise Mismatch(other, "", "directory is empty")

    def __call__(self, **kwargs):
        return self.__class__(**kwargs)


IsDirectory = _IsDirectory()



import pytest
from majava import DictContains, Any
from majava.formats import IsJson


@pytest.mark.parametrize("actual, expected", [
    ("[1, 2, 3]", IsJson()),
    ("[1, 2, 3]", IsJson([1, 2, 3])),
    ('{"a": 1, "b": 2}', IsJson({"a": 1, "b": 2})),
    (
        '{"a": 1, "b": 2}',
        IsJson(DictContains({"b": 2}))
    ),
    (
        '''{
            "a": "[1, 2, 3.14, 4]",
            "b": "any value may be here",
            "c": {"x": 1, "y": 2}
        }''',
        IsJson({
            "a": IsJson([1, 2, Any, 4]),
            "b": Any,
            "c": DictContains({"x": 1})
        })
    )
])
def test_is_json__match(actual, expected):
    assert actual == expected


@pytest.mark.parametrize("actual, expected, reason", [
    (None, {}, (
        "Value None at 'FromJSON' does not match: invalid type - "
        "the JSON object must be str, bytes or bytearray, not NoneType"
    )),
    ("", {}, (
        "Value '' at 'FromJSON' does not match: invalid JSON - "
        "Expecting value: line 1 column 1 (char 0)"
    )),
    (
        '''{
            "a": "[1, 2, 3.14, 4.14]",
            "b": "any value may be here",
            "c": {"x": 1, "y": 2}
        }''',
        {
            "a": IsJson([1, 2, Any, 4]),
            "b": Any,
            "c": DictContains({"x": 1})
        },
        (
            "Value [1, 2, 3.14, 4.14] at 'FromJSON.a.FromJSON' does not match: "
            "[1, 2, 3.14, 4.14] != [1, 2, <Any>, 4]'"
        )
    )
])
def test_is_json__mismatch(actual, expected, reason):
    with pytest.raises(AssertionError) as e:
        assert actual == IsJson(expected)

    err_message = str(e.value)
    reason = err_message.splitlines()[-1].strip()
    assert reason == reason



from .matchers import Matcher, And, Or, Any, MayBe, Absent, matcher
from .basic import IsInstance, DictContains, Round, Contains, Unordered, \
    StartsWith, EndsWith, LengthIs, InInterval, HasAttrs


__all__ = [
    "Matcher", "And", "Or", "Any", "MayBe", "Absent", "matcher",
    "InInterval", "IsInstance", "DictContains", "Round", "Contains", "Unordered",
    "StartsWith", "EndsWith", "LengthIs", "HasAttrs"
]



from typing import Optional, Type, Callable
import inspect


class Mismatch(Exception):
    @classmethod
    def invalid_type(cls, value, expected_type, path=""):
        return cls(value, path, f"invalid type - got {type(value)}, expected {expected_type}")

    @classmethod
    def invalid_len(cls, value, expected_len, path=""):
        return cls(value, path, f"invalid length - got {len(value)}, expected {expected_len}")

    @classmethod
    def missing_keys(cls, value, keys, path=""):
        keys_str = ", ".join(repr(i) for i in keys)
        return cls(value, path, f"missing items with keys: {keys_str}")

    @classmethod
    def unexpected_keys(cls, value, keys, path=""):
        keys_str = ", ".join(repr(i) for i in keys)
        raise Mismatch(value, "", f"unexpected items with keys: {keys_str}")

    @classmethod
    def missing_items(cls, value, items, path=""):
        items_str = ", ".join(repr(i) for i in items)
        raise Mismatch(value, "", f"missing items: {items_str}")

    @classmethod
    def missing_item(cls, value, path=""):
        return cls(value, path, "missing item")

    @classmethod
    def unexpected_item(cls, value, path=""):
        return cls(value, path, "unexpected item")

    def __init__(self, value, path, msg):
        self.value = value
        self.path = path
        self.msg = msg

    def prepend(self, path):
        self.path = f"{path}.{self.path}" if self.path else path
        return self

    def __str__(self):
        if not self.path:
            return f"Value {repr(self.value)} does not match: {self.msg}"
        return f"Value {repr(self.value)} at {repr(self.path)} does not match: {self.msg}"


class Matcher:
    """ Base class for all matchers.
    """

    _mismatch = None

    def __eq__(self, other):
        try:
            self._match(other)
            self._mismatch = None
            return True
        except Mismatch as e:
            self._mismatch = e
            return False

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def _match(self, other) -> Optional[str]:
        pass


def make_matcher(fn: Callable) -> Type[Matcher]:
    """ Decorates a function to become a matcher.
    """

    name = fn.__name__

    argspec = inspect.getfullargspec(fn)
    arg_count = len(argspec.args) - 1

    class M(Matcher):
        def __init__(self, *args, **kwargs):
            if len(args) != arg_count:
                raise TypeError(
                    f"{name}() takes {arg_count} positional arguments, {len(args)} were given")
            if argspec.varkw is None and kwargs:
                raise TypeError(
                    f"{name}() got an unexpected keyword argument '{next(iter(kwargs))}'")
            self.args = args
            self.kwargs = kwargs

        def __repr__(self):
            args_str = ", ".join(repr(i) for i in self.args)
            kwargs_str = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
            content_str = ", ".join(filter(None, [args_str, kwargs_str]))
            return f"{name}({content_str})"

        def _match(self, other):
            if fn(other, *self.args, **self.kwargs) is False:
                raise Mismatch(other, "", f"not {self}")

    M.__qualname__ = name
    M.__name__ = name
    M.__doc__ = fn.__doc__

    return M


class _MatcherWrap(Matcher):
    def __init__(self, v):
        self.v = v

    def __repr__(self):
        return repr(self.v)

    def _match(self, other):
        _match(self.v, other)


def matcher(value) -> Matcher:
    """ Makes a matcher from the given value.
    It allows to get similar message on AssertionError in pytest.
    """

    return _MatcherWrap(value)


def _check_type(value, types):
    if not isinstance(value, types):
        raise Mismatch.invalid_type(value, types)


def _check_len(value, expected):
    v_len = len(value)
    if v_len != expected:
        raise Mismatch(value, "", f"invalid length - got {v_len}, expected {expected}")


def _match(matcher, value):
    if isinstance(matcher, Matcher):
        return matcher._match(value)

    if isinstance(matcher, dict):
        _check_type(value, dict)
        return _match_dict(matcher, value)

    if matcher == value:
        return
    if matcher is Absent:
        raise Mismatch.unexpected_item(value)
    if value is Absent:
        raise Mismatch.missing_item(value)
    raise Mismatch(value, "", f"{repr(value)} != {repr(matcher)}")


class And(Matcher):
    def __init__(self, *matchers, repr=None):
        self.matchers = matchers
        self._repr = repr

    def __and__(self, other):
        return And(*self.matchers, other)

    def __repr__(self):
        if self._repr is not None:
            return self._repr
        return '&'.join(repr(it) for it in self.matchers)

    def _match(self, other):
        for matcher in self.matchers:
            _match(matcher, other)


class Or(Matcher):
    def __init__(self, *matchers):
        self.matchers = matchers

    def __or__(self, other):
        return Or(*self.matchers, other)

    def __repr__(self):
        return '|'.join(repr(it) for it in self.matchers)

    def _match(self, other):
        mismatches = []
        for matcher in self.matchers:
            try:
                _match(matcher, other)
                return
            except Mismatch as e:
                mismatches.append(e)

        or_str = " nor ".join(repr(i) for i in self.matchers)
        raise Mismatch(other, "", f"is not {or_str}")


class _Any:
    def __repr__(self):
        return "<Any>"

    def __eq__(self, other):
        return other is not Absent


class _Absent:
    def __repr__(self):
        return "<Absent>"

    def __eq__(self, other):
        return other is self


Any = _Any()
Absent = _Absent()


class MayBe(Matcher):
    """ To be used in containers. Item may not exist or must match.
    """

    def __init__(self, v):
        self.v = v

    def __repr__(self):
        return f"MayBe({repr(self.v)})"

    def _match(self, other):
        if self is Absent:
            return
        _match(self.v, other)


def _is_missing(val):
    return not isinstance(val, (MayBe, _Absent))


def _match_dict(matcher: dict, value: dict, allow_unexpected=False):
    missing_keys = set(matcher.keys())
    unexpected_keys = []

    for key, value_v in value.items():
        try:
            matcher_v = matcher[key]
        except KeyError:
            if not allow_unexpected:
                unexpected_keys.append(key)
            continue

        if matcher_v is Absent:
            unexpected_keys.append(key)
            continue

        try:
            _match(matcher_v, value_v)
        except Mismatch as e:
            raise e.prepend(key)

        missing_keys.remove(key)

    missing_keys = sorted(filter(lambda k: _is_missing(matcher[k]), missing_keys))
    if missing_keys:
        raise Mismatch.missing_keys(value, missing_keys)

    if unexpected_keys:
        raise Mismatch.unexpected_keys(value, unexpected_keys)



from majava import InInterval, IsInstance, Round, StartsWith, EndsWith, LengthIs, HasAttrs, \
    Any, Absent
from .common import raises_assertion_error


def test_is_instance():
    m = IsInstance(int, str)

    assert str(m) == "IsInstance(int|str)"
    assert 1 == m
    assert "2" == m
    with raises_assertion_error("Value 1.1 does not match: not IsInstance(int|str)"):
        assert 1.1 == m


def test_round():
    assert repr(Round(3.14)) == "Round(3.0)"
    assert repr(Round(3.14, 2)) == "Round(3.14)"

    assert 3.0 == Round(3.14)
    assert 3.14 == Round(3.14)
    assert 3.144 == Round(3.14, 2)

    with raises_assertion_error("Value 3.146 does not match: not ~3.14"):
        assert 3.146 == Round(3.14, 2)


def test_starts_with():
    m = StartsWith("no")

    assert repr(m) == "StartsWith('no')"
    assert "no money" == m
    assert "no niin" == m
    with raises_assertion_error("Value 'Honey' does not match: not StartsWith('no')"):
        assert "Honey" == m


def test_ends_with():
    m = EndsWith("yes")

    assert repr(m) == "EndsWith('yes')"
    assert "yes" == m
    assert "told yes" == m
    with raises_assertion_error("Value 'yes my lord' does not match: not EndsWith('yes')"):
        assert "yes my lord" == m


def test_in_interval():
    m = InInterval(1, 7)

    assert repr(m) == "InInterval(1, 7)"
    assert 1 == m
    assert 4 == m
    assert 7 == m
    with raises_assertion_error("Value 10 does not match: not InInterval(1, 7)"):
        assert 10 == m


def test_length_is():
    m = LengthIs(3)

    assert repr(m) == "LengthIs(3)"
    assert [1, 2, 3] == m
    assert "abc" == m
    with raises_assertion_error("Value 'abcd' does not match: not LengthIs(3)"):
        assert "abcd" == m


def test_has_attrs():
    class AttrDict(dict):
        def __getattr__(self, key):
            if key not in self:
                raise AttributeError(key)
            return self[key]

    m = HasAttrs(a=1, b=Any, c=Absent)

    assert repr(m) == "HasAttrs(a=1, b=<Any>, c=<Absent>)"
    assert AttrDict({"a": 1, "b": 2}) == m
    assert AttrDict({"a": 1, "b": None, "x": 3}) == m

    with raises_assertion_error("Value 2 at 'a' does not match: 2 != 1"):
        assert AttrDict({"a": 2, "b": 2}) == m
    with raises_assertion_error("Value <Absent> at 'b' does not match: missing item"):
        assert AttrDict({"a": 1}) == m
    with raises_assertion_error("Value 3 at 'c' does not match: unexpected item"):
        assert AttrDict({"a": 1, "b": 2, "c": 3}) == m



import pytest
from .matchers import Matcher


def pytest_assertrepr_compare(config: pytest.Config, op, left, right):
    if not (isinstance(left, Matcher) or isinstance(right, Matcher)):
        return

    matcher = left if isinstance(left, Matcher) else right

    return [
        f"{left} {op} {right}",
        f"{matcher._mismatch}",
    ]
